
from __future__ import annotations

import argparse
from dataclasses import dataclass, field, fields
import logging
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Type, TypeVar

from .errors import ConfigError

# TODO - Python 3.11 will have public logging.getLevelNamesMapping
# https://github.com/python/cpython/issues/88024
# There is also a solution for click: https://pypi.org/project/click-loglevel/
log_levels = tuple(p[1] for p in sorted(logging._levelToName.items()))

# click and typer are good, but other lint tools like black and mypy depend on
# click (as does typer). Projects that want to use omnidep might also be using
# any past or future version of black/mypy/etc. I don't want to risk
# incompatibilities due to breaking changes in a common dependency, or do much
# work on an upgrade treadmill ensuring I always support every version of
# click. So, argparse it is.
parser = argparse.ArgumentParser(description="Check project dependencies against imports in code.")

CLT = TypeVar('CLT', bound='BasicCommandLine')

@dataclass
class BasicCommandLine:
    _log_level: Optional[str] = None
    verbose: bool = False

    @property
    def log_level(self) -> Optional[str]:
        if self.verbose:
            return 'INFO'
        return self._log_level

    @classmethod
    def parse(cls: Type[CLT], args: Optional[List[str]] = None) -> CLT:
        return parser.parse_args(args=args, namespace=cls())

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument('--log-level', choices=log_levels, dest='_log_level')
        parser.add_argument('--verbose', '-v', action='store_true', default=False)

@dataclass
class CommandLine(BasicCommandLine):
    paths: List[Path] = field(default_factory=list)
    project: Optional[Path] = None
    tests: Optional[List[Path]] = None

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument('paths', metavar='PATH', nargs='+', type=Path)
        parser.add_argument('--project', metavar='PATH', type=Path)
        parser.add_argument('--tests', metavar='PATH', action='append', type=Path)
        super().add_arguments(parser)

CommandLine.add_arguments(parser)

@dataclass(frozen=True)
class Config:
    ignore_imports: List[str] = field(default_factory=list)
    child_packages: Dict[str, List[str]] = field(default_factory=dict)
    ignore_dependencies: List[str] = field(default_factory=list)
    local_test_packages: List[str] = field(default_factory=list)
    local_test_paths: List[Path] = field(default_factory=list)
    ignore_dependencies_order: bool = False
    ignore_dev_dependencies_order: bool = False

    @classmethod
    def make(cls, data: Optional[Dict[str, Any]] = None) -> Config:
        data = data or {}
        allowed = {f.name: f for f in fields(cls)}
        args: Dict[str, Any] = {}
        for orig_key, value in data.items():
            key = orig_key.replace('-', '_')
            # Warn for typos or misunderstandings. Also rejects options from
            # future versions of the code. That's OK because the same project
            # file containing the options can require a version of omnidep that
            # supports them.
            if key not in allowed:
                raise ConfigError(f"Config option {orig_key!r} not recognised")
            # Try to give a useful error message for misconfigurations.
            # Not many types needed, so just hack them rather than add a
            # dependency on pydantic.
            check: Dict[str, Callable[[object], bool]]
            check = {
                'bool': lambda val: isinstance(val, bool),
                'List[str]': lambda val: (
                    isinstance(val, list) and
                    all(isinstance(x, str) for x in val)
                ),
                'List[Path]': lambda val: (
                    isinstance(val, list) and
                    all(isinstance(x, str) for x in val)
                ),
                'Dict[str, List[str]]': lambda val: (
                    isinstance(val, dict) and
                    all(isinstance(key, str) for key in val) and
                    all(map(check['List[str]'], val.values()))
                ),
            }
            convert: Dict[str, Callable[[Any], Any]] = {
                'List[Path]': lambda val: list(map(Path, val))
            }
            # It already is a str due to "import annotations". But mypy doesn't
            # seem to know that, so coerce it.
            expected_type = str(allowed[key].type)
            if not check[expected_type](value):
                raise ConfigError(f"Config option {orig_key!r}: expected {expected_type}, got {value!r}")
            converter = convert.get(expected_type, lambda x: x)
            args[key] = converter(value)
        return Config(**args)



#!/usr/bin/env python3
#

import logging
from pathlib import Path
from typing import Iterable, Optional

from .command import CommandLine
from .errors import ConfigError
from .project import read_poetry

logger = logging.getLogger()


def get_project_file(paths: Iterable[Path]) -> Optional[Path]:
    tomls = set(path for path in paths if path.name == 'pyproject.toml')
    if len(tomls) > 1:
        raise SystemExit("ERROR: Multiple pyproject.toml files specified")
    if len(tomls) == 1:
        return tomls.pop()
    return None

def main(args: CommandLine) -> None:
    warnings = (
        read_poetry(args.project or get_project_file(args.paths))
        .collect(lambda x: x.check_dependencies(args.paths, exclude=args.tests or ()))
        .collect(lambda x: x.check_dev_dependencies(args.tests))
    ).warnings

    if warnings:
        print('\n'.join(w.report for w in warnings))
        print("See https://github.com/sjjessop/omnidep#error-codes-explained")
    else:
        logger.info("No issues found")

def script_entry_point() -> None:
    args = CommandLine.parse()
    if args.log_level is not None:
        logging.basicConfig(level=args.log_level)
    try:
        main(args)
    except ConfigError as e:
        raise SystemExit(str(e))
    finally:
        print("")

if __name__ == '__main__':
    script_entry_point()




from pathlib import Path
from typing import List

import pytest

from omnidep import project
from omnidep.errors import Violation

test_dir = Path(__file__).parent
root_dir = test_dir.parent.parent

def test_self() -> None:
    """Must be able to load self as a project"""
    assert project.read_poetry(root_dir / 'pyproject.toml').warnings == ()


plain_project_files = [
    (test_dir / 'test_cases/unsorted', [Violation.ODEP006]),
    (test_dir / 'test_cases/case_sensitive_sorted', [Violation.ODEP006]),
    (test_dir / 'test_cases/case_insensitive_sorted', []),
    (test_dir / 'test_cases/dev_dependencies_old', [Violation.ODEP007]),
    (test_dir / 'test_cases/dev_dependencies_new', [Violation.ODEP007]),
]

@pytest.mark.parametrize('dir,expected', plain_project_files)
def test_known_project_files(dir: Path, expected: List[Violation]) -> None:
    """
    Must generate the expected warnings from known pyproject.toml files

    These test cases don't include any source code, just the toml.
    """
    warnings = project.read_poetry(dir / 'pyproject.toml').warnings
    assert [warning.code for warning in warnings] == expected




from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, unique
import itertools
from typing import Callable, Generic, Iterable, Optional, Tuple, TypeVar

class ConfigError(ValueError):
    pass

@unique
class Violation(Enum):
    # Might do a flake8 plugin some day, so use suitable codes
    ODEP001 = 'missing-dependency'
    ODEP002 = 'module-not-found'
    ODEP003 = 'namespace-used'
    ODEP004 = 'missing-namespace-dependency'
    ODEP005 = 'unused-dependency'
    ODEP006 = 'unsorted-dependency'
    ODEP007 = 'uncanonical-name'
    ODEP008 = 'unmanaged-module'

    def __call__(self, msg: str, package: Optional[str] = None) -> Warn:
        return Warn(self, msg, package)

@dataclass(frozen=True)
class Warn:
    code: Violation
    msg: str
    missing_package_name: Optional[str]

    @property
    def report(self) -> str:
        return f"{self.code.name}: {self.msg}"

T = TypeVar('T', covariant=True)
U = TypeVar('U')

@dataclass(frozen=True)
class Warned(Generic[T]):
    """
    Represents a value, plus zero or more warnings related to the generation of
    that value.

    This is more-or-less a Writer monad, where the semigroup (over which the
    written data is summed) is fixed as an iterable with concatenation.
    """
    value: T
    # TODO - tuple not efficient. Linked list?
    warnings: Tuple[Warn, ...] = ()

    def flatMap(self, func: Callable[[T], Warned[U]]) -> Warned[U]:
        """Change the value and append any number of warnings"""
        result = func(self.value)
        return Warned(result.value, self.warnings + result.warnings)
    def map(self, func: Callable[[T], U]) -> Warned[U]:
        """Change the value without changing the warnings"""
        return self.set(func(self.value))
    def collect(self, func: Callable[[T], Iterable[Warn]]) -> Warned[T]:
        """Append any number of warnings generated from the value, without changing the value"""
        return Warned(self.value, self.warnings + tuple(func(self.value)))

    def set(self, value: U) -> Warned[U]:
        """Set a new value without changing the warnings"""
        return Warned(value, self.warnings)

    def warn(self, warning: Warn) -> Warned[T]:
        """Append one warning without changing the value"""
        # This is like the function tell of cats.Writer
        return Warned(self.value, self.warnings + (warning,))
    def warnAll(self, warnings: Iterable[Warn]) -> Warned[T]:
        """Append any number of warnings without changing the value"""
        return Warned(self.value, self.warnings + tuple(warnings))

    def as_tuple(self) -> Tuple[T, Tuple[Warn, ...]]:
        """Return (value, warnings)"""
        return self.value, self.warnings
    @staticmethod
    def gather(items: Iterable[Warned[T]]) -> Warned[Tuple[T, ...]]:
        """
        Combine multiple Warned objects in the order specified.

        The values are placed into a tuple, and all warnings are concatenated.
        """
        # It's fine to read the iterable all at once, since zip will anyway.
        items = tuple(items)
        if len(items) == 0:
            # zip doesn't handle the case of no items.
            return Warned(())
        values, warningses = zip(*map(Warned.as_tuple, items))
        return Warned(values, tuple(itertools.chain.from_iterable(warningses)))




from pathlib import Path
import sys

import pytest

from omnidep import packages

first_names = ('foo', 'Foo', 'FOO')
last_names = ('bar', 'Bar', 'BAR')
middle_bits = ('-', '_', '.', '__', '-_', '...')

@pytest.mark.parametrize('first', first_names)
@pytest.mark.parametrize('middle', middle_bits)
@pytest.mark.parametrize('last', last_names)
def test_canon(first: str, middle: str, last: str) -> None:
    assert packages.canon(first + middle + last) == 'foo-bar'

@pytest.mark.parametrize('first', first_names)
@pytest.mark.parametrize('middle', middle_bits)
@pytest.mark.parametrize('last', last_names)
@pytest.mark.parametrize('suffix', middle_bits)
def test_canon_suffix(first: str, middle: str, last: str, suffix: str) -> None:
    assert packages.canon(first + middle + last + suffix) == 'foo-bar-'

def test_get_preferred_name() -> None:
    assert packages.get_preferred_name('non-existent-package') is None
    assert packages.get_preferred_name('pytest') == 'pytest'
    assert packages.get_preferred_name('Pytest') == 'pytest'
    assert packages.get_preferred_name('PYTEST') == 'pytest'
    assert packages.get_preferred_name('pyOpenSSL') == 'pyOpenSSL'
    assert packages.get_preferred_name('pyopenssl') == 'pyOpenSSL'

def test_find_packages() -> None:
    local = frozenset({'omnidep'})

    # Non-existent package is found nowhere, represented by empty list
    result = packages.find_packages('non-existent-package', local)
    assert result.value == []
    assert result.warnings == ()

    # Installed package must be found
    result = packages.find_packages('omnidep', local)
    assert result.value == ['omnidep']
    assert result.warnings == ()

    # TODO - this is what currently happens, but it's wrong
    result = packages.find_packages('OmniDep', local)
    assert result.value == ['OmniDep']
    assert result.warnings == ()

    # Case where the imported name doesn't match the distributed name.
    # This is dev dependency only, although it'll be pulled in by all sorts.
    result = packages.find_packages('OpenSSL', local)
    assert result.value == ['pyOpenSSL']
    assert result.warnings == ()

    # Find packages on the path (with a warning)
    sys.path.append(str(Path(__file__).parent.parent))
    try:
        result = packages.find_packages('tst', frozenset())
        assert result.value == ['tst']
        assert len(result.warnings) == 1
    finally:
        sys.path.pop()
    # Check it wouldn't have been found without the path
    result = packages.find_packages('tst', frozenset())
    assert result.value == []
    assert result.warnings == ()




from __future__ import annotations

from dataclasses import dataclass
import itertools
import logging
from pathlib import Path
from typing import (
    Any, Collection, Container, Dict, FrozenSet, Iterable, Optional, Set,
    Tuple,
)

import tomli

from .command import Config
from .errors import Violation as V
from .errors import Warn, Warned
from .imports import find_source_files, get_external_modules
from .packages import canon, find_packages, get_preferred_name

logger = logging.getLogger()


def check_order(deps: Collection[str], label: str = 'dependencies') -> Iterable[Warn]:
    deps = list(deps)
    if not deps:
        return
    if deps[0] == 'python':
        deps = deps[1:]
    for first, second in zip(deps, itertools.islice(deps, 1, None)):
        if first.casefold() > second.casefold():
            yield V.ODEP006(f"{label} are not sorted: {first!r} before {second!r}")
            return

def fix_canonical_names(data: Dict[str, Any]) -> Warned[FrozenSet[str]]:
    def check_canon(package_name: str) -> Warned[str]:
        canonical_name = canon(package_name)
        if package_name != canonical_name:
            # The preferred name of the package (like "Django") might not be
            # the same as the canonical form ("django"). We allow the user to
            # specify either of the two, and warn for any other form.
            preferred_name = get_preferred_name(package_name) or canonical_name
            if package_name != preferred_name:
                return Warned(canonical_name).warn(
                    V.ODEP007(f"dependency {package_name!r} is not the preferred name: consider {preferred_name!r}")
                )
        return Warned(canonical_name)
    return Warned.gather(map(check_canon, data)).map(frozenset)

@dataclass(frozen=True)
class Project:
    dependencies: Collection[str]
    dev_dependencies: Collection[str]
    config: Config = Config.make(None)
    local_packages: FrozenSet[str] = frozenset()
    extra_paths: Tuple[Path, ...] = ()

    def check_dependencies(
        self, paths: Iterable[Path],
        *, exclude: Iterable[Path] = (),
    ) -> Iterable[Warn]:
        yield from self.check_modules(
            itertools.chain(paths, self.extra_paths),
            self.dependencies,
            self.local_packages,
            exclude=itertools.chain(exclude, self.config.local_test_paths),
        )

    def check_dev_dependencies(self, paths: Optional[Iterable[Path]]) -> Iterable[Warn]:
        if paths is None:
            return
        yield from self.check_modules(
            paths,
            self.dev_dependencies,
            self.local_packages | set(self.config.local_test_packages),
            label='dev-dependencies',
            # Because dev-dependencies includes linters etc. that aren't used
            # anywhere in the code.
            check_unused=False,
        )

    def check_modules(
        self, paths: Iterable[Path], packages: Collection[str], local_packages: FrozenSet[str],
        *, label: str = 'dependencies', check_unused: bool = True, exclude: Iterable[Path] = (),
    ) -> Iterable[Warn]:
        paths = list(paths)
        logger.info(f'searching {", ".join(map(str, paths))}')
        def get_files(paths: Iterable[Path]) -> Set[Path]:
            return set(itertools.chain.from_iterable(map(find_source_files, paths)))
        included_paths = get_files(paths) - get_files(exclude)
        modules = [
            module for module in get_external_modules(included_paths)
            if not self.ignore_import(module)
        ]
        logger.info(f'{label} imported: {modules}')
        used: Set[str] = {'python'}
        for module in modules:
            founds = find_packages(module, local_packages)
            yield from founds.warnings
            found = list(map(canon, founds.value))
            if len(found) == 1:
                package = found[0]
                used.add(package)
                if package not in local_packages and not self.required(package, packages):
                    yield V.ODEP001(f'Package {package!r} is imported but not listed in {label}', package)
            elif len(found) == 0:
                yield V.ODEP002(f"Module {module!r} is imported but not installed, so I don't know what package is needed", module)
            else:
                # Namespace package - implemented across multiple installed
                # packages. So for any given import, we don't know which
                # package supplies that part of the namespace package.
                options = [pkg in local_packages or self.required(pkg, packages) for pkg in found]
                if all(options):
                    # OK, if we depend on them all that's fine
                    used.update(found)
                elif any(options):
                    yield V.ODEP003(f"Namespace package found: any of {found} might provide {module!r}")
                    used.update(found)
                else:
                    yield V.ODEP004(f"Namespace package found: any of {found} might provide {module!r}, and there are no dependencies on any of them", module)
        if check_unused:
            unused = set(packages) - used - set(self.config.ignore_dependencies)
            if unused:
                yield V.ODEP005(f'Unused {label} in project file: {unused}')

    def ignore_import(self, module: str) -> bool:
        return bool(self.config) and module in self.config.ignore_imports

    def required(self, package: str, packages: Container[str]) -> bool:
        def mentioned(pkg: str) -> bool:
            return pkg in packages or pkg in self.local_packages
        if mentioned(package):
            return True
        for parent, children in self.config.child_packages.items():
            if package in children and mentioned(parent):
                return True
        return False

def read_poetry(toml_file: Optional[Path]) -> Warned[Project]:
    if toml_file is None:
        logger.error("pyproject.toml not specified")
        return Warned(Project((), ()))
    with open(toml_file, 'rb') as infile:
        tools = tomli.load(infile)['tool']
    poetry_data = tools['poetry']
    config = Config.make(tools.get('omnidep'))

    def process(deps: Dict[str, Any], dev: bool) -> Warned[FrozenSet[str]]:
        if dev:
            ignore = config.ignore_dev_dependencies_order
            key = 'dev-dependencies'
        else:
            ignore = config.ignore_dependencies_order
            key = 'dependencies'
        return (
            Warned(deps)
            .collect(lambda x: () if ignore else check_order(x, key))
            .flatMap(fix_canonical_names)
        )

    deps = process(poetry_data['dependencies'], False)
    # Poetry 1.2.0+ has two different places you can specify dev dependencies
    old_dev_data = poetry_data.get('dev-dependencies', {})
    old_dev_deps = process(old_dev_data, True)
    new_dev_data = poetry_data.get('group', {}).get('dev', {}).get('dependencies', {})
    new_dev_deps = process(new_dev_data, True)
    dev_deps: Warned[FrozenSet[str]] = (
        Warned.gather([old_dev_deps, new_dev_deps])
        .map(lambda x: frozenset(itertools.chain.from_iterable(x)))
    )

    pkgs = {pack['include'] for pack in poetry_data['packages']}
    project = Project(
        dependencies=deps.value,
        dev_dependencies=deps.value | dev_deps.value,
        config=config,
        local_packages=frozenset(map(canon, pkgs)),
        extra_paths=tuple(toml_file.parent / pack for pack in pkgs),
    )
    return Warned.gather([deps, dev_deps]).set(project)




from pathlib import Path
from unittest import mock

from omnidep import imports

test_dir = Path(__file__).parent

def test_find_source_files() -> None:
    results = list(imports.find_source_files(test_dir))
    assert Path(__file__) in results
    assert all(path.suffix == '.py' for path in results)

def test_find_source_files_single() -> None:
    this_file = Path(__file__)
    assert list(imports.find_source_files(this_file)) == [Path(this_file)]

def test_is_external() -> None:
    assert not imports.is_external('pathlib')
    assert not imports.is_external('__future__')
    assert imports.is_external('omnidep')
    assert imports.is_external('pytest')
    # Not strictly true, but for our purposes
    assert not imports.is_external('setuptools')

def test_get_external_modules() -> None:
    results = imports.get_external_modules([test_dir])
    assert 'pathlib' not in results
    assert 'unittest' not in results
    assert 'omnidep' in results

def test_every_import() -> None:
    test_file = test_dir / 'test_cases' / 'every_import.py~'
    with mock.patch('omnidep.imports.find_source_files') as find:
        find.return_value = [test_file]
        results = list(imports.get_external_modules([Path('.')]))
    # Must not think 'bad' is a top-level import, in any place it occurs
    assert 'bad' not in results
    # Must find all the expected top-level imports, once each
    expected_results = sorted(f'example{x}' for x in range(1, 25))
    assert results == expected_results




from typing import Callable, Iterable, Tuple

import pytest

from omnidep.errors import Violation, Warn, Warned

def assert_odep1(msg: str, warning: Warn) -> None:
    assert warning.code == Violation.ODEP001
    assert warning.msg == msg
    assert warning.missing_package_name is None
    assert warning.report == "ODEP001: " + msg

def test_violation() -> None:
    """Warnings are constructed from Violations"""
    warning = Violation.ODEP001("foo")
    assert_odep1("foo", warning)
    assert_odep1("bar", Violation.ODEP001("bar"))
    assert_odep1("baz", Violation.ODEP001("baz"))

def test_warned_constructor_1() -> None:
    """Single-arg constructor of Warned"""
    result = Warned(1)
    assert result.value == 1
    assert list(result.warnings) == []
    assert result.as_tuple() == (1, ())

def test_warned_constructor_2() -> None:
    """Two-arg constructor of Warned, with empty initial warnings"""
    result = Warned(2, ())
    assert result.value == 2
    assert list(result.warnings) == []
    assert result.as_tuple() == (2, ())

def test_warned_constructor_warnings() -> None:
    """Two-arg constructor of Warned, with a warning"""
    result = Warned(3, (Violation.ODEP001("foo"),))
    assert result.value == 3
    warnings = list(result.warnings)
    assert len(warnings) == 1
    assert_odep1("foo", warnings[0])
    assert result.as_tuple() == (3, (Violation.ODEP001("foo"),))

def test_warned_equality() -> None:
    """All the other tests will be based on equality, so it better be right."""
    def equal(m1: Warned[int], m2: Warned[int]) -> None:
        assert m1 == m2
        assert not m1 != m2
        assert m1.as_tuple() == m2.as_tuple()
        assert m1.value == m2.value
        assert list(m1.warnings) == list(m2.warnings)
    def notequal(m1: Warned[int], m2: Warned[int]) -> None:
        assert m1 != m2
        assert not m1 == m2
        assert m1.as_tuple() != m2.as_tuple()
        assert m1.value != m2.value or list(m1.warnings) != list(m2.warnings)
    for v in range(3):
        equal(Warned(v), Warned(v, ()))
    warns = [(), (warning1,), (warning1, warning2)]
    cases = [Warned(v, w) for v in range(3) for w in warns]
    for m in cases:
        equal(m, m)
    for left in range(len(cases)):
        for right in range(len(cases)):
            if left != right:
                notequal(cases[left], cases[right])

###########################
# Functions to test flatMap
###########################

def bad(x: int) -> Tuple[Warn, ...]:
    # Warning message depends on x value, so when we chain together to test
    # associativity we're asserting that the message order matches.
    return Violation.ODEP001(f"bad {x}"), Violation.ODEP001("and another thing")

def unit(x: int) -> Warned[int]:
    return Warned(x)
def unit_bad(x: int) -> Warned[int]:
    return Warned(x, bad(x))
def increment(x: int) -> Warned[int]:
    return Warned(x + 1)
def increment_bad(x: int) -> Warned[int]:
    return Warned(x + 1, bad(x))
def double(x: int) -> Warned[int]:
    return Warned(x * 2)
def double_bad(x: int) -> Warned[int]:
    return Warned(x * 2, bad(x))

all_funcs = [unit, unit_bad, increment, increment_bad, double, double_bad]
bad_funcs = [f for f in all_funcs if f.__name__.endswith('_bad')]
good_funcs = [f for f in all_funcs if not f.__name__.endswith('_bad')]
all_values = [0, 1, 2, 3, 19, 2**128 - 1]

IntFunc = Callable[[int], Warned[int]]


###########################################################
# flatMap applies the functions to produce expected results
###########################################################

@pytest.mark.parametrize('value', all_values)
@pytest.mark.parametrize('f', [unit, unit_bad])
def test_noop(value: int, f: IntFunc) -> None:
    assert Warned(value).flatMap(f).value == value

@pytest.mark.parametrize('value', all_values)
@pytest.mark.parametrize('f', [increment, increment_bad])
def test_increment(value: int, f: IntFunc) -> None:
    assert Warned(value).flatMap(f).value == value + 1

@pytest.mark.parametrize('value', all_values)
@pytest.mark.parametrize('f', [double, double_bad])
def test_double(value: int, f: IntFunc) -> None:
    assert Warned(value).flatMap(f).value == value * 2

@pytest.mark.parametrize('value', all_values)
@pytest.mark.parametrize('f', bad_funcs)
def test_msg(value: int, f: IntFunc) -> None:
    result = Warned(value).flatMap(f)
    assert list(result.warnings) == [*bad(value)]

@pytest.mark.parametrize('value', all_values)
@pytest.mark.parametrize('f', bad_funcs)
def test_msgs(value: int, f: IntFunc) -> None:
    result = Warned(value).flatMap(f).flatMap(f)
    second_value = f(value).value
    assert list(result.warnings) == [*bad(value), *bad(second_value)]

@pytest.mark.parametrize('value', all_values)
@pytest.mark.parametrize('f', good_funcs)
def test_no_msgs(value: int, f: IntFunc) -> None:
    result = Warned(value).flatMap(f)
    assert list(result.warnings) == []
    result = result.flatMap(f)
    assert list(result.warnings) == []
    result = result.flatMap(f)
    assert list(result.warnings) == []


############
# Monad laws
############

@pytest.mark.parametrize('value', all_values)
@pytest.mark.parametrize('f', all_funcs)
def test_monad_law_left(value: int, f: IntFunc) -> None:
    """
    Left-identity

    Applying unit, then flatmap, then f, is the same as just f.
    """
    assert Warned(value).flatMap(f) == f(value)

@pytest.mark.parametrize('value', all_values)
def test_monad_law_right(value: int) -> None:
    """
    Right-identity

    .flatMap(unit) does nothing.
    """
    for m in (Warned(value), Warned(value).flatMap(double_bad)):
        # Two ways of expressing the unit. We can't just say Warned because
        # it accepts anything and deduces type, hence has the wrong signature.
        for f in (unit, lambda x: Warned(x)):
            assert m.flatMap(f) == m

@pytest.mark.parametrize('value', all_values)
@pytest.mark.parametrize('f', all_funcs)
@pytest.mark.parametrize('g', all_funcs)
def test_monad_law_assoc(value: int, f: IntFunc, g: IntFunc) -> None:
    """
    Associativity

    flatMap f then g is the same as flatMap the composition.
    """
    for m in (Warned(value), Warned(value).flatMap(double_bad)):
        assert m.flatMap(f).flatMap(g) == m.flatMap(lambda x: f(x).flatMap(g))


####################################################
# gather(), which distributes Iterable across Warned
####################################################

def test_gather() -> None:
    warneds = [
        Warned(1),
        Warned(2).flatMap(double_bad),
        Warned(3).flatMap(double_bad).flatMap(double_bad),
    ]
    result: Warned[Tuple[int, ...]] = Warned.gather(iter(warneds))
    assert result.value == (1, 4, 12)
    assert list(result.warnings) == [*bad(2), *bad(3), *bad(6)]

def test_gather_empty() -> None:
    result: Warned[Tuple[int, ...]] = Warned.gather([])
    assert result.value == ()
    assert list(result.warnings) == []


##################################
# Additional convenience functions
#
# Each test has an assertion that
# defines the convenience function
# in terms of flatMap.
##################################

warning1 = Violation.ODEP001("one")
warning2 = Violation.ODEP002("two")
warning3 = Violation.ODEP003("three")

@pytest.mark.parametrize('value', all_values)
@pytest.mark.parametrize('f', all_funcs)
def test_map(value: int, f: IntFunc) -> None:
    """map() applies a "dropped" function that doesn't return Warned"""
    def g(x: int) -> int:
        return f(x).value
    result = Warned(value).map(g)
    assert result.value == Warned(value).flatMap(f).value
    assert result.value == g(value)
    assert list(result.warnings) == []
    for m in (Warned(value), Warned(value).map(g)):
        assert m.map(g) == m.flatMap(lambda x: Warned(g(x)))

@pytest.mark.parametrize('value', all_values)
def test_warn(value: int) -> None:
    """warn() adds a single warning"""
    result = Warned(value).warn(warning1).warn(warning2)
    assert result.value == value
    assert list(result.warnings) == [warning1, warning2]
    for m in (Warned(value), Warned(value).warn(warning1)):
        assert m.warn(warning2) == m.flatMap(lambda x: Warned(x, (warning2,)))

@pytest.mark.parametrize('value', all_values)
def test_warnAll(value: int) -> None:
    """warnAll() adds multiple warnings"""
    result = Warned(value).warnAll([warning1]).warnAll([warning2, warning3])
    assert result.value == value
    assert list(result.warnings) == [warning1, warning2, warning3]
    for m in (Warned(value), Warned(value).warnAll([warning1])):
        assert m.warnAll([warning2, warning3]) == m.flatMap(lambda x: Warned(x, (warning2, warning3)))

@pytest.mark.parametrize('value', all_values)
def test_collect(value: int) -> None:
    """
    collect() also adds multiple warnings, but they're generated from the
    current value.
    """
    def gen_warnings(value: int) -> Iterable[Warn]:
        return [*bad(value)]
    result = Warned(value).collect(gen_warnings)
    assert result.value == value
    assert list(result.warnings) == [*bad(value)]
    for m in (Warned(value), result):
        assert m.collect(gen_warnings) == m.flatMap(lambda x: Warned(x, tuple(gen_warnings(x))))

@pytest.mark.parametrize('x', all_values)
@pytest.mark.parametrize('y', all_values)
def test_set(x: int, y: int) -> None:
    """set() just changes the value"""
    assert Warned(x).set(y) == Warned(y)
    assert Warned(x, (warning1, warning2)).set(y) == Warned(y, (warning1, warning2))
    for m in (Warned(x), Warned(x, (warning1, warning2))):
        assert m.set(y) == m.flatMap(lambda x: Warned(y))




import collections
import contextlib
import functools
from pathlib import Path
import re
import sys
from typing import FrozenSet, List, Mapping, Optional

import importlib_metadata as metadata

from .errors import Violation as V
from .errors import Warned

punctuation = re.compile(r'[\-._]+')

# In Python 3.9+, should use functools.cache instead of lru_cache
# In Python 3.10+, there is metadata.packages_distributions, but all it checks
# is top_level.txt, so we still need to search for files as well.
@functools.lru_cache()
def packages_distributions() -> Mapping[str, List[str]]:
    pkg_to_dist = collections.defaultdict(set)
    for dist in metadata.distributions():
        def add_to(pkg: str) -> None:
            pkg_to_dist[pkg].add(dist.metadata['Name'])
        for toplevel in (dist.read_text('top_level.txt') or '').split():
            add_to(toplevel)
        for file in dist.files or ():
            # TODO - maybe the package could contain .pyc or .pyd but no .py
            if file.name == '__init__.py':
                add_to(str(file.parent))
            elif str(file.parent) == '.' and file.suffix == '.py':
                add_to(file.stem)
    # TODO - make the return immutable, since it's cached
    return {key: sorted(value) for key, value in pkg_to_dist.items()}

def find_packages(module: str, local_packages: FrozenSet[str]) -> Warned[List[str]]:
    """
    Given a top-level code module, which installed package(s) provide it?
    This is a difficult question because Python packaging doesn't try to fully
    answer it, hence we need to apply some guesswork.
    """
    if canon(module) in local_packages:
        return Warned([module])
    # TODO - Perhaps a more sure way would be to import the module and then
    # look for __file__ in all the packages.files
    #
    # If a package lists our module in its top-level.txt or sources, it will
    # appear here.
    package = packages_distributions().get(module)
    if package is not None:
        return Warned(list(package))
    # Maybe the package is on the path, in which case no package dependency is
    # needed provided that it remains available on the path.
    if any((Path(path) / module).is_dir() for path in sys.path):
        return Warned([module]).warn(
            V.ODEP008(f'Module {module!r} not under package management but found on python path')
        )
    return Warned([])

def canon(package_name: str) -> str:
    """
    Return the normalized name per PEP 503:
    https://peps.python.org/pep-0503/#normalized-names
    """
    return '-'.join(punctuation.split(package_name)).lower()

def get_preferred_name(package: str) -> Optional[str]:
    """
    Return the name the project calls itself.
    """
    # importlib_metadata.PackageNotFoundError inherits from FileNotFoundError
    # in old versions (<3) and ImportError more recently.
    with contextlib.suppress(FileNotFoundError, ImportError):
        name: Optional[str] = metadata.distribution(package).metadata.get('Name')
        return name
    return None




import ast
import itertools
from pathlib import Path
import sys
from typing import Iterable, List

basic_types = (str, float, int, bytes, type(...))

def iter_import_names(tree: ast.AST) -> Iterable[str]:
    to_process: List[object] = [tree]
    while to_process:
        node = to_process.pop()
        if isinstance(node, ast.ImportFrom):
            # print(ast.unparse(node))
            if node.level == 0 and node.module is not None:
                yield node.module.partition('.')[0]
        elif isinstance(node, ast.Import):
            # print(ast.unparse(node))
            for alias in node.names:
                yield alias.name.partition('.')[0]
        elif node is None:
            pass
        elif isinstance(node, list):
            to_process.extend(node)
        elif isinstance(node, ast.AST):
            to_process.extend(getattr(node, name) for name in node._fields)
        elif not isinstance(node, basic_types):
            raise NotImplementedError(f'unhandled {type(node)} {node!r}')

def find_source_files(path: Path) -> Iterable[Path]:
    if path.is_file() and path.suffix == '.py':
        return [path]
    return path.glob('**/*.py')

def iter_modules(path: Path) -> Iterable[str]:
    for file in find_source_files(path):
        with open(file, encoding='utf8') as infile:
            # print(file)
            tree = ast.parse(infile.read())
            yield from iter_import_names(tree)

def is_external(module: str) -> bool:
    if module in ('setuptools', 'pkg_resources'):
        # Debateable: not technically part of Python, but distributed with it.
        return False
    if sys.version_info >= (3, 10):
        if module == '__future__':
            return False
        return module not in sys.stdlib_module_names
    else:
        from isort import place_module
        return str(place_module(module)) not in ('STDLIB', 'FUTURE')

def get_external_modules(paths: Iterable[Path]) -> List[str]:
    all_modules = itertools.chain.from_iterable(map(iter_modules, paths))
    return sorted(set(filter(is_external, all_modules)))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['omnidep', 'omnidep.tst']

package_data = \
{'': ['*'],
 'omnidep.tst': ['test_cases/*',
                 'test_cases/case_insensitive_sorted/*',
                 'test_cases/case_sensitive_sorted/*',
                 'test_cases/dev_dependencies_new/*',
                 'test_cases/dev_dependencies_old/*',
                 'test_cases/unsorted/*']}

install_requires = \
['importlib-metadata>=1.1.0', 'tomli>=1.1.0']

extras_require = \
{':python_version < "3.10"': ['isort>=5.0.1']}

entry_points = \
{'console_scripts': ['omnidep = omnidep.main:script_entry_point']}

setup_kwargs = {
    'name': 'omnidep',
    'version': '0.3.2',
    'description': 'Linter to compare project dependencies against imports in source code',
    'long_description': '=======\nOmnidep\n=======\n\nA Python linter to compare a project\'s declared dependencies against the import\nstatements in its source code.\n\n.. image:: https://github.com/sjjessop/omnidep/workflows/tests/badge.svg?branch=develop\n   :alt: Test status\n   :target: https://github.com/sjjessop/omnidep/actions?query=workflow%3Atests+branch%3Adevelop\n\n.. image:: https://img.shields.io/badge/CI%20python-3.7%20%7C%203.8%20%7C%203.9%20%7C%203.10%20%7C%203.11-blue.svg\n   :alt: Tested with Python versions 3.7 3.8 3.9 3.10 3.11\n   :target: https://www.python.org/downloads/\n\n.. image:: https://img.shields.io/pypi/pyversions/omnidep\n   :alt: PyPI project\n   :target: https://pypi.org/project/omnidep/\n\n.. image:: https://img.shields.io/badge/badges-4-green.svg\n   :alt: 4 badges\n   :target: https://shields.io/\n\nPurpose\n=======\n\nProvides warnings when a project imports packages that it doesn\'t declare a\ndependency on, plus some related linting of the project dependency data.\n\nCurrently only poetry projects are supported (configured in pyproject.toml).\nThe projects that your project depends on can be packaged using any tools, but\nyour project (that omnidep analyses) currently must use poetry.\n\nInstallation\n------------\n\n.. code-block:: bash\n\n    pip install omnidep\n\nOptionally, you can also run the test suite. This would be a good idea if\nyou\'re using a new (or pre-release) version of Python not included in this\nrepo\'s CI testing.\n\n.. code-block:: bash\n\n    pip install pyOpenSSL pytest\n    pytest --pyargs omnidep.tst\n\nUsage\n-----\n\n.. code-block:: bash\n\n    omnidep pyproject.toml\n\n\nConfiguration\n-------------\n\nomnidep uses your project\'s poetry configuration to work out:\n\n* What source files to search for imports, from ``tool.poetry.packages``.\n* What dependencies your project declares, from ``tool.poetry.dependencies``.\n* What dev-dependencies your project declares, from\n  ``tool.poetry.dev-dependencies`` and ``tool.poetry.group.dev.dependences``.\n\nIf you have test code that you want omnidep to search for imports, then:\n\n* If you keep your test code "inside" your project, then list it in\n  ``local-test-paths`` in the ``[omnidep.config]`` section described below.\n* If you keep your test code "outside" your project, then use the ``--tests``\n  command-line option to locate it, otherwise omnidep ignores it. You also need\n  to configure ``local-test-packages`` if some of your test files import other\n  test files, for example if you have shared helper functions.\n\nomnidep is configured using the ``pyproject.toml`` file, and specifically the\n``[tool.omnidep]`` section. The following config keys are recognised.\nUnrecognised keys are rejected and omnidep will not run (so, if you want to\nuse a particular key then you should require at least the minimum version of\nomnidep that recognises it).\n\nignore-imports\n^^^^^^^^^^^^^^\n\nExample: ``ignore-imports = ["X"]``\n\nSince: 0.2.0\n\nCauses omnidep to ignore all import statements from X, for example\n``import X``, ``from X.Y import Z``. omnidep will behave as if your code does\nnot use package X, even if it does. X must be a top-level package. It is not\ncurrently possible to selectively ignore a sub-package (like X.Y), nor is it\ncurrently possible to ignore imports from some files but not others.\n\nchild-packages\n^^^^^^^^^^^^^^\n\nExample: ``child-packages = {boto3 = ["botocore"]}``\n\nSince: 0.2.0\n\nCauses omnidep to consider a dependency on boto3 to also supply botocore. This\nsaves you having to explictly list the child as a dependency of your project.\nYou chould only do this when the child is inherant to the parent, not just\nbecause by chance you pull in a package you need via an indirect dependency.\nThe reason is that indirect dependencies can change, and the project that you\ndo depend on might not require the same version of the child that your usage\nrequires. Only if the projects are closely related can you assume that the\nversion you require of one will provide the features you need from the other.\n\nignore-dependencies\n^^^^^^^^^^^^^^^^^^^\n\nExample: ``ignore-dependencies = ["X"]``\n\nSince: 0.2.0\n\nCauses omnidep to ignore the project X listed in your project\'s dependencies.\nomnidep will behave as if your project does not depend on X, even if it does.\n\nignore-dependencies-order\n^^^^^^^^^^^^^^^^^^^^^^^^^\n\nExample: ``ignore-dependencies-order = true``\n\nSince: 0.2.0\n\nCauses omnidep to skip the check that your dependencies are alphabetically\nordered.\n\nignore-dev-dependencies-order\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nExample: ``ignore-dev-dependencies-order = true``\n\nSince: 0.2.0\n\nCauses omnidep to skip the check that your dev-dependencies are alphabetically\nordered.\n\nlocal-test-paths\n^^^^^^^^^^^^^^^^\n\nExample: ``local-test-paths = ["myproject/tests/"]``\n\nSince: 0.2.0\n\nCauses omnidep to treat all code in ``myproject.tests`` as test code, meaning\nthat anything it imports can be provided either by your projects dependencies\nor by its dev-dependencies. Imports from code that is not test code must be\nprovided by non-dev dependencies.\n\nlocal-test-packages\n^^^^^^^^^^^^^^^^^^^\n\nExample: ``local-test-packages = ["tests"]``\n\nSince: 0.2.0\n\nCauses omnidep to treat ``tests`` as part of the current project, but only when\nconsidering imports that appear in test code. Use this when your test code is\nnot shipped as part of your project.\n\nError codes explained\n---------------------\n\nX, Y, P, Q, R, represent the names of imports or dependencies, depending on the\nmessage.\n\nODEP001\n^^^^^^^\n\n| ``package \'X\' is imported but not listed in dependencies``\n| ``package \'X\' is imported but not listed in dev-dependencies``\n|\n\nX is the name you imported, which is not necessarily the same as the name of\nthe project you have to install (for example the project ``beautifulsoup4``\ninstalls the package ``bs4``). omnidep does its best to find what project your\ndesired package comes from, but if it fails, or if you don\'t have a suitable\ndependency, then this is the result.\n\nTo fix, choose one of the following:\n\n* List the project name in your dependencies. If the package is used from test\n  code, then the dependency can be either dev or non-dev. If the package is\n  used from non-test code, then the dependency needs to be non-dev.\n* To ignore the import, add it to the list of ignored imports in your\n  ``[tool.omnidep]`` config, like ``ignore-imports = ["X"]``.\n* The package might come from a dependency of a dependency, and you might\n  prefer not to explicitly list it as a direct dependency too, so you can list\n  X as a child of some other dependency that you do list. You should only do\n  this when the indirect dependency is inherent to the direct dependency, for\n  example ``boto3`` provides ``botocore``. Add\n  ``child-packages = {something = ["X"]}`` to your ``[tool.omnidep]`` config,\n  meaning that the project named "something" provides "X", and so a dependency\n  on "something" is acceptable in place of a dependency on "X".\n\nODEP002\n^^^^^^^\n\n``module \'X\' is imported but not installed``\n\nNot only is there no dependency found that provides X, but X isn\'t even\ncurrently installed. omnidep relies on locally installed metadata to help it\nfind what dependencies correspond to what imports.\n\nTo fix, choose one of the following:\n\n* If your project has X as a dependency, but you haven\'t installed your\n  project then install your project, bringing in its dependencies.\n* Add a dependency that provides X.\n* Ignore the import by listing it in your in your ``[tool.omnidep]`` config,\n  like ``ignore-imports = ["X"]``.\n\nODEP003\n^^^^^^^\n\n``Namespace package found: any of [\'P\', \'Q\', \'R\'] might provide \'X\'``\n\nIf projects P, Q, and R all provide code in the Python package X, then omnidep\ndoesn\'t know which one you need in order to satisfy a given import. If you\ndeclare dependencies on all of them (that is, all the ones you currently have\ninstalled), then omnidep is satisfied. If you depend on some but not others,\nthen you get this message.\n\nTo fix, choose one of the following:\n\n* If you don\'t need the ones you don\'t declare dependencies on, and they are\n  installed accidentally, then uninstall them.\n* If appropriate, declare dependencies on all of P, Q, and R. However, this\n  might not be appropriate because P and Q might be genuine direct dependencies\n  of your code, whereas R was pulled in indirectly via something else. You\n  don\'t want to have to list indirect dependencies as direct dependencies.\n* Otherwise you have to resolve for yourself whether your dependencies are\n  adequate, then ignore the import with ``ignore-imports = ["X"]`` in your\n  ``[tool.omnidep]`` config.\n\n\nODEP004\n^^^^^^^\n\n``Namespace package found: any of [\'P\', \'Q\', \'R\'] might provide \'X\', and there are no dependencies on any of them``\n\nIf projects P, Q, and R all provide code in the Python package X, then omnidep\ndoesn\'t know which one you need in order to satisfy a given import. If you\ndeclare dependencies on all of them (that is, all the ones you currently have\ninstalled), then omnidep is satisfied. If you depend on none of them,\nthen you get this message.\n\nTo fix, choose one of the following:\n\n* If appropriate, declare dependencies on all of P, Q, and R. However, this\n  might not be appropriate because P and Q might be genuine direct dependencies\n  of your code, whereas R was pulled in indirectly via something else. You\n  don\'t want to have to list indirect dependencies as direct dependencies.\n* Otherwise you have to resolve for yourself whether your dependencies are\n  adequate, then ignore the import with ``ignore-imports = ["X"]`` in your\n  ``[tool.omnidep]`` config.\n\n\nODEP005\n^^^^^^^\n\n``unused dependencies in project file: {\'X\', \'Y\'}``\n\nomnidep expects you not to list any dependencies that you don\'t import. This\nmight be completely legitimate, for example:\n\n* the dependency is a plugin to some framework and will be used via some means\n  other than an explicit ``import`` in your code;\n* you are controlling the version of an indirect dependency, to deal with\n  some problem caused by unexpected breaking changes.\n\nUnused dev-dependencies are always ignored, since they tend to include linters\nand suchlike.\n\nTo fix, choose one of the following:\n\n* Remove the dependency.\n* List the dependency in your ``[tool.omnidep]`` config like\n  ``ignore-dependencies = ["X"]``.\n\nODEP006\n^^^^^^^\n\n| ``dependencies are not sorted: \'Y\' before \'X\'``\n| ``dev-dependencies are not sorted: \'Y\' before \'X\'``\n|\n\nIgnoring ``python``, which is allowed to come first, omnidep expects you to\nlist dependencies in case-insensitive alphabetical order within each section\n(dev and non-dev).\n\nTo fix, choose one of the following:\n\n* List your dependencies alphabetically.\n* Set ``ignore-dependencies-order = true`` or\n  ``ignore-dev-dependencies-order = true`` in your ``[tool.omnidep]`` config.\n\nODEP007\n^^^^^^^\n\n``dependency \'X\' is not the preferred name: consider \'Y\'``\n\nomnidep expects you to use either of two formats to name dependencies in your\nproject file: the "Normalized Name" as defined in\n`PEP 503 <https://peps.python.org/pep-0503/>`_ or the name the dependency uses\nfor itself in its metadata. Any name that normalizes to the same value will\nwork, but inconsistent naming tends to lead to confusion, or to failing to find\nmentions when you search for them.\n\nTo fix:\n\n* Use the name omnidep suggests, or the normalized name.\n\nODEP008\n^^^^^^^\n\n``Module \'X\' not under package management but found on python path``\n\nomnidep cannot find any project that provides X, but it is available to import.\nThis can happen for example if you have set up the ``PYTHONPATH`` to find the\ncode, instead of installing it as a dependency.\n\nTo fix, choose one of the following:\n\n* If this is an error, list a suitable dependency.\n* If you know what you\'re doing, and users of your project will know how to\n  supply the code that you\'re importing, then ignore the import with\n  ``ignore-imports = ["X"]`` in your ``[tool.omnidep]`` config.\n\nChangelog\n=========\n\n0.3.2\n-----\n\n* Read dev dependencies from the new location used by Poetry 1.2.0+\n* Add Python 3.11 to PyPI classifiers.\n\n0.3.1\n-----\n\n* Add Python 3.11 to the test matrix, and use separate badges in the README\n  for what is tagged on PyPI vs. what is tested.\n* Documentation improvements.\n* Uncap Python dependency. If Python ever reaches version 4, you are free to\n  install omnidep on it and see what happens!\n\n0.3.0\n-----\n\n* Breaking: When testing that dependencies are sorted, do it case-insensitive.\n* Deal with some build issues.\n\n0.2.1\n-----\n\n* Refer to online docs insted of long message in terminal.\n* Publish to PyPI.\n\n0.2.0\n-----\n\n* Minor documentation improvements.\n* Lower bounds for dependencies importlib-metadata, isort, and tomli.\n* CI test of the lower-bound versions.\n',
    'author': 'Steve Jessop',
    'author_email': '68118527+sjjessop@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sjjessop/omnidep',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0',
}


setup(**setup_kwargs)
